"""Tests the service layer."""

import logging
from pathlib import Path
from textwrap import dedent
from typing import Optional

import pytest

from yamlfix import fix_files
from yamlfix.model import YamlfixConfig, YamlNodeStyle
from yamlfix.services import fix_code

true_strings = [
    "TRUE",
    "True",
    "true",
    "YES",
    "Yes",
    "yes",
    "ON",
    "On",
    "on",
]

false_strings = [
    "FALSE",
    "False",
    "false",
    "NO",
    "No",
    "no",
    "OFF",
    "Off",
    "off",
]


class TestFixFiles:
    """Test the fix_files function."""

    def test_fix_files_can_process_string_arguments(self, tmp_path: Path) -> None:
        """
        Given: A file to fix
        When: Passing the string with the path to the file to fix_files
        Then: The file is fixed
        """
        test_file = tmp_path / "source.yaml"
        test_file.write_text("program: yamlfix")
        fixed_source = dedent(
            """\
            ---
            program: yamlfix
            """
        )

        fix_files([str(test_file)], False)  # act

        assert test_file.read_text() == fixed_source

    def test_fix_files_issues_warning(self, tmp_path: Path) -> None:
        """
        Given: A file to fix
        When: Using the old signature
        Then: A warning is issued
        """
        test_file = tmp_path / "source.yaml"
        test_file.write_text("program: yamlfix")

        with pytest.warns(UserWarning, match="yamlfix/pull/182"):
            fix_files([str(test_file)])


class TestFixCode:
    """Test the fix_code function."""

    def test_fix_code_ignore_jinja2(self) -> None:
        """Ignores jinja2 line if present at the beginning of the source."""
        source = dedent(
            """\
            # jinja2:lstrip_blocks: true
            ---
            program: yamlfix
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_ignore_shebang(self) -> None:
        """Ignores shebang lines if present at the beginning of the source."""
        source = dedent(
            """\
            #! /this/line/should/be/ignored
            ---
            program: yamlfix
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_ignore_ansible_vaults(self) -> None:
        """Adds the --- at the beginning of the source."""
        source = dedent(
            """\
            $ANSIBLE_VAULT;1.1;AES256
            3036303361343731386530393763...
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_adds_header(self) -> None:
        """Adds the --- at the beginning of the source."""
        source = "program: yamlfix"
        fixed_source = dedent(
            """\
            ---
            program: yamlfix
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    def test_fix_code_doesnt_double_the_header(self) -> None:
        """If source starts with --- don't add another line."""
        source = dedent(
            """\
            ---
            program: yamlfix
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_corrects_indentation_on_lists(self) -> None:
        """Use two spaces for indentation of lists."""
        source = dedent(
            """\
            ---
            hosts:
            - item1
            - item2
            """
        )
        fixed_source = dedent(
            """\
            ---
            hosts:
              - item1
              - item2
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.KEEP_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_fix_code_respects_parent_lists(self) -> None:
        """Do not indent lists at the first level."""
        source = dedent(
            """\
            ---
            - item1
            - item2
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_preserves_comments(self) -> None:
        """Don't delete comments in the code."""
        source = dedent(
            """\
            ---
            # Keep comments!
            program: yamlfix
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_respects_parent_lists_with_comments(self) -> None:
        """Do not indent lists at the first level even if there is a comment."""
        source = dedent(
            """\
            ---
            # Comment
            - item1
            - item2
            """
        )

        result = fix_code(source)

        assert result == source

    @pytest.mark.parametrize(
        "code",
        [
            dedent(
                """\
                ---
                - program:
                  # Keep comments!
                """
            ),
            dedent(
                """\
                ---
                - name: Setup SONiC Build Servers
                  hosts: sonic_build_server
                  vars:
                    # Keep comments!
                    build_user: build
                """
            ),
            dedent(
                """\
                ---
                tasks:
                  - name: Make sure repository is cloned  # noqa latest[git]
            """
            ),
        ],
    )
    def test_fix_code_preserves_indented_comments(self, code: str) -> None:
        """Don't remove indentation from comments in the code."""
        result = fix_code(code)

        assert result == code

    def test_fix_code_removes_extra_apostrophes(self) -> None:
        """Remove not needed apostrophes."""
        source = dedent(
            """\
            ---
            title: 'Why we sleep'
            """
        )
        fixed_source = dedent(
            """\
            ---
            title: Why we sleep
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    @pytest.mark.parametrize("true_string", true_strings)
    def test_fix_code_converts_non_valid_true_booleans(self, true_string: str) -> None:
        """Convert common strings that refer to true, but aren't the string `true`.

        [More
        info](https://yamllint.readthedocs.io/en/stable/rules.html#module-yamllint.rules.truthy)
        """
        source = dedent(
            f"""\
            ---
            True dictionary: {true_string}
            True list:
            - {true_string}
            """
        )
        fixed_source = dedent(
            """\
            ---
            True dictionary: true
            True list: [true]
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    @pytest.mark.parametrize("false_string", false_strings)
    def test_fix_code_converts_non_valid_false_booleans(
        self, false_string: str
    ) -> None:
        """Convert common strings that refer to false, but aren't the string `false`.

        [More
        info](https://yamllint.readthedocs.io/en/stable/rules.html#module-yamllint.rules.truthy)
        """
        source = dedent(
            f"""\
            ---
            False dictionary: {false_string}
            False list:
            - {false_string}
            """
        )
        fixed_source = dedent(
            """\
            ---
            False dictionary: false
            False list: [false]
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    @pytest.mark.parametrize("truthy_string", true_strings + false_strings)
    def test_fix_code_respects_apostrophes_for_truthy_substitutions(
        self,
        truthy_string: str,
    ) -> None:
        """Keep apostrophes for strings like `yes` or `true`.

        So they are not converted to booleans.
        """
        source = dedent(
            f"""\
            ---
            title: '{truthy_string}'
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_adds_space_in_comment(self) -> None:
        """Correct comments that don't have a space between
        the # and the first character.
        """
        source = dedent(
            """\
            ---
            #This is a comment
            project: yamlfix
            """
        )
        fixed_source = dedent(
            """\
            ---
            # This is a comment
            project: yamlfix
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    def test_fix_code_not_add_extra_space_in_comment(self) -> None:
        """Respects comments that already have a space between
        the # and the first character.
        """
        source = dedent(
            """\
            ---
            # This is a comment
            project: yamlfix
            """
        )
        fixed_source = dedent(
            """\
            ---
            # This is a comment
            project: yamlfix
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    def test_fix_code_add_space_inline_comment(self) -> None:
        """Fix inline comments that don't have a space between
        the # and the first character.
        """
        source = dedent(
            """\
            ---
            project: yamlfix  #This is a comment
            """
        )
        fixed_source = dedent(
            """\
            ---
            project: yamlfix  # This is a comment
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    def test_fix_code_respects_url_anchors(self) -> None:
        """Comments that contain a url with an anchor should be respected."""
        source = dedent(
            """\
            ---
            # https://lyz-code.github.io/yamlfix/#usage
            foo: bar
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_add_extra_space_inline_comment(self) -> None:
        """Fix inline comments that don't have two spaces before
        the #.
        """
        source = dedent(
            """\
            ---
            project: yamlfix # This is a comment
            """
        )
        fixed_source = dedent(
            """\
            ---
            project: yamlfix  # This is a comment
            """
        )

        result = fix_code(source)

        assert result == fixed_source

    def test_fix_code_doesnt_change_double_exclamation_marks(self) -> None:
        """Lines with starting double exclamation marks should be respected, otherwise
        some programs like mkdocs-mermaidjs fail.
        """
        source = dedent(
            """\
            ---
            format: !!python/name:mermaid2.fence_mermaid
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_parses_files_with_multiple_documents(self) -> None:
        """Files that contain multiple documents should be parsed as a collection of
        separate documents and then dumped together again.
        """
        source = dedent(
            """\
            ---
            project: yamlfix
            ---
            project: yamlfix
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_functions_emit_debug_logs(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Each fixer function should emit a log at the debug level in each run."""
        caplog.set_level(logging.DEBUG)

        fix_code("")  # act

        expected_logs = {
            "Setting up ruamel yaml 'quote simple values' configuration...",
            "Setting up ruamel yaml 'sequence flow style' configuration...",
            "Running ruamel yaml base configuration...",
            "Running source code fixers...",
            "Fixing truthy strings...",
            "Fixing jinja2 variables...",
            "Running ruamel yaml fixer...",
            "Restoring truthy strings...",
            "Restoring jinja2 variables...",
            "Restoring double exclamations...",
            "Fixing comments...",
            "Fixing top level lists...",
            "Fixing flow-style lists...",
        }
        assert set(caplog.messages) == expected_logs
        for record in caplog.records:
            assert record.levelname == "DEBUG"

    @pytest.mark.parametrize("whitespace", ["", "\n", "\n\n"])
    def test_fixed_code_has_exactly_one_newline_at_end_of_file(
        self,
        whitespace: str,
    ) -> None:
        """Files should have exactly one newline at the end to comply with the POSIX
        standard.
        """
        source = dedent(
            """\
            ---
            program: yamlfix"""
        )
        source += whitespace
        fixed_code = dedent(
            """\
            ---
            program: yamlfix
            """
        )

        result = fix_code(source)

        assert result == fixed_code

    def test_anchors_and_aliases_with_duplicate_merge_keys(self) -> None:
        """All anchors and aliases should be preserved even with multiple merge keys
        and merge keys should be formatted as a list in a single line.
        """
        source = dedent(
            """\
            ---
            x-node-volumes: &node-volumes
              node3_data:
            x-vault-volumes: &vault-volumes
              vault_data:
            x-mongo-volumes: &mongo-volumes
              mongo_data:
            x-certmgr-volumes: &certmgr-volumes
              cert_data:
            volumes:
              <<: *node-volumes
              <<: *vault-volumes
              <<: *mongo-volumes
              <<: *certmgr-volumes
            """
        )
        desired_source = dedent(
            """\
            ---
            x-node-volumes: &node-volumes
              node3_data:
            x-vault-volumes: &vault-volumes
              vault_data:
            x-mongo-volumes: &mongo-volumes
              mongo_data:
            x-certmgr-volumes: &certmgr-volumes
              cert_data:
            volumes:
              <<:
                - *node-volumes
                - *vault-volumes
                - *mongo-volumes
                - *certmgr-volumes
            """
        )
        config = YamlfixConfig()
        config.allow_duplicate_keys = True

        result = fix_code(source, config)

        assert result == desired_source

    def test_fix_code_respects_comment_symbol_in_strings_with_simple_quotes(
        self,
    ) -> None:
        """
        Given: Code with a string that contains a #
        When: fix_code is run
        Then: The string is left unchanged
        """
        source = dedent(
            """\
            ---
            project: 'Here # is not a comment marker'
            """
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_respects_comment_symbol_in_strings_with_double_quotes(
        self,
    ) -> None:
        """
        Given: Code with a string that contains a #
        When: fix_code is run
        Then: The string is left unchanged
        """
        source = dedent(
            """\
            ---
            project: "Here # is not a comment marker"
            """
        )
        desired_source = dedent(
            """\
            ---
            project: 'Here # is not a comment marker'
            """
        )

        result = fix_code(source)

        assert result == desired_source

    def test_fix_code_respects_jinja_variables_with_equals(
        self,
    ) -> None:
        """
        Given: Code with a long string that contains a jinja variable after an equal
        When: fix_code is run
        Then: The jinja string is not broken
        """
        source = (
            "---\n"
            "environment:\n"
            "  - SUPER_LONG_VARIABLE={{ this_is_a_super_long_long_long_long_long"
            "_variable_name }}\n"
        )

        result = fix_code(source)

        assert result == source

    def test_fix_code_respects_many_jinja_variables(
        self,
    ) -> None:
        """
        Given: Code with a long string that contains two jinja variables
        When: fix_code is run
        Then: The jinja string is not broken
        """
        source = (
            "---\n"
            "project: This is a long long {{ variable_1 }} line that should not be "
            "split on the jinja {{ variable_2 }}"
        )
        desired_source = (
            "---\n"
            "project: This is a long long {{ variable_1 }} line that should not be "
            "split on the\n"
            "  jinja {{ variable_2 }}\n"
        )

        result = fix_code(source)

        assert result == desired_source

    def test_fix_code_respects_jinja_variables_with_operations(
        self,
    ) -> None:
        """
        Given: Code with a long string that contains a jinja variable with operations
        When: fix_code is run
        Then: The jinja string is not broken
        """
        source = (
            "---\n"
            "project: This is a long long long long line that should not be split on "
            "the jinja {{ variable that contains different words }}"
        )
        desired_source = (
            "---\n"
            "project: This is a long long long long line that should not be split on "
            "the jinja\n"
            "  {{ variable that contains different words }}\n"
        )

        result = fix_code(source)

        assert result == desired_source

    def test_fix_code_doesnt_remove_variable(
        self,
    ) -> None:
        """
        Given: Code with a short jinja variable
        When: fix_code is run
        Then: The jinja string is not removed
        """
        source = "---\npath: '{{ item }}'\n"

        result = fix_code(source)

        assert result == source

    @pytest.mark.parametrize(
        ("source", "config", "desired_source"),
        [
            (
                dedent(
                    """\
                    ---
                    list: [item, item]


                    # Comment: desired: 1 line (default) before this comment
                    key: value
                    """
                ),
                None,
                dedent(
                    """\
                    ---
                    list: [item, item]

                    # Comment: desired: 1 line (default) before this comment
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]



                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=1),
                dedent(
                    """\
                    ---
                    list: [item, item]
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]



                    key: value  # Comment: desired: No lines between `list` and `key`
                    """
                ),
                YamlfixConfig(comments_whitelines=0),
                dedent(
                    """\
                    ---
                    list: [item, item]
                    key: value  # Comment: desired: No lines between `list` and `key`
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]



                    key: value  # Comment: desired: No lines between `list` and `key`
                    """
                ),
                YamlfixConfig(comments_whitelines=1),
                dedent(
                    """\
                    ---
                    list: [item, item]
                    key: value  # Comment: desired: No lines between `list` and `key`
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]



                    key: value  # Comment: desired: No lines between `list` and `key`
                    """
                ),
                YamlfixConfig(comments_whitelines=2),
                dedent(
                    """\
                    ---
                    list: [item, item]
                    key: value  # Comment: desired: No lines between `list` and `key`
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]


                    # Comment: desired: 1 line before this comment
                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=1),
                dedent(
                    """\
                    ---
                    list: [item, item]

                    # Comment: desired: 1 line before this comment
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]


                    # Comment: desired: 0 line before this comment
                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=0),
                dedent(
                    """\
                    ---
                    list: [item, item]
                    # Comment: desired: 0 line before this comment
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]


                    # Comment: desired: 2 lines before this comment
                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=2),
                dedent(
                    """\
                    ---
                    list: [item, item]


                    # Comment: desired: 2 lines before this comment
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list: [item, item]

                    # Comment: desired: 2 lines before this comment
                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=2),
                dedent(
                    """\
                    ---
                    list: [item, item]


                    # Comment: desired: 2 lines before this comment
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list:
                      - item
                      # Comment: desired: 0 new line before this comment
                      - item

                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=1),
                dedent(
                    """\
                    ---
                    list:
                      - item
                      # Comment: desired: 0 new line before this comment
                      - item
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list:
                      - item
                      # Comment: desired: 0 new line before this comment
                      - item

                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=0),
                dedent(
                    """\
                    ---
                    list:
                      - item
                      # Comment: desired: 0 new line before this comment
                      - item
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list:
                      - item
                      # Comment: desired: 0 new line before this comment
                      - item

                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=2),
                dedent(
                    """\
                    ---
                    list:
                      - item
                      # Comment: desired: 0 new line before this comment
                      - item
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list:
                      - item

                      # Comment: desired: 1 new line before this comment
                      - item

                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=1),
                dedent(
                    """\
                    ---
                    list:
                      - item

                      # Comment: desired: 1 new line before this comment
                      - item
                    key: value
                    """
                ),
            ),
            (
                dedent(
                    """\
                    ---
                    list:
                      - item

                      # Comment: desired: 2 new lines before this comment
                      - item

                    key: value
                    """
                ),
                YamlfixConfig(comments_whitelines=2),
                dedent(
                    """\
                    ---
                    list:
                      - item


                      # Comment: desired: 2 new lines before this comment
                      - item
                    key: value
                    """
                ),
            ),
        ],
    )
    def test_fix_code_fix_whitelines(
        self,
        source: str,
        config: Optional[YamlfixConfig],
        desired_source: str,
    ) -> None:
        """
        Given: Code with extra whitelines
        When: fix_code is run
        Then: The string has extra whitelines removed
        """
        result = fix_code(source_code=source, config=config)

        assert result == desired_source



"""Test the Yaml and YamlRoundTrip adapters."""

from textwrap import dedent

import pytest
from ruyaml.constructor import DuplicateKeyError

from yamlfix.model import YamlfixConfig, YamlNodeStyle
from yamlfix.services import fix_code

none_representations = [
    "",
    "null",
    "Null",
    "NULL",
    "~",
]

quote_representations = [
    "'",
    '"',
]


class TestYamlAdapter:
    """Test the Yaml and YamlRoundTrip adapters."""

    def test_indentation_config(self) -> None:
        """Make indentation values configurable."""
        source = dedent(
            """\
            project_name: yamlfix
            list:
              - item
            map:
              key: value
            """
        )
        fixed_source = dedent(
            """\
            ---
            project_name: yamlfix
            list:
                -   item
            map:
                key: value
            """
        )
        config = YamlfixConfig()
        config.indent_offset = 4
        config.indent_mapping = 4
        config.indent_sequence = 8
        config.sequence_style = YamlNodeStyle.KEEP_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_dont_allow_duplicate_keys_config(self) -> None:
        """Test if duplicate keys cause an exception when configured."""
        source = dedent(
            """\
            ---
            project_name: yamlfix
            project_name: yumlfax
            """
        )
        config = YamlfixConfig()
        config.allow_duplicate_keys = False

        with pytest.raises(
            DuplicateKeyError, match='found duplicate key "project_name"'
        ):
            fix_code(source, config)

    def test_comment_spacing_config(self) -> None:
        """Test if spaces are added to comment start if configured."""
        source = dedent(
            """\
            ---
            # comment
            project_name: yamlfix #comment
            """
        )
        fixed_source = dedent(
            """\
            ---
            # comment
            project_name: yamlfix  # comment
            """
        )
        config = YamlfixConfig()
        config.comments_min_spaces_from_content = 2
        config.comments_require_starting_space = True

        result = fix_code(source, config)

        assert result == fixed_source

    def test_dont_generate_explicit_start(self) -> None:
        """Test if the explicit yaml document start indicator is removed\
            when configured."""
        source = dedent(
            """\
            ---
            project_name: yamlfix
            """
        )
        fixed_source = dedent(
            """\
            project_name: yamlfix
            """
        )
        config = YamlfixConfig()
        config.explicit_start = False

        result = fix_code(source, config)

        assert result == fixed_source

    def test_if_line_length_expands(self) -> None:
        """Test if configurable line-length expands string value."""
        source = dedent(
            """\
            key: value value value value value value
              value value value value value value
              value value value value value value
              value value value value value value
              value value value value value value
              value value value value value value
            """
        )
        fixed_source = dedent(
            """\
            ---
            key: value value value value value value value value value value value value value value value value value
              value value value value value value value value value value value value value value value value value
              value value
            """  # noqa: E501
        )
        config = YamlfixConfig()
        config.line_length = 100

        result = fix_code(source, config)

        assert result == fixed_source

    def test_if_line_length_contracts(self) -> None:
        """Test if configurable line-length contracts string value."""
        source = dedent(
            """\
            key: value value value value value value
              value value value value value value
              value value value value value value
              value value value value value value
              value value value value value value
              value value value value value value
            """
        )
        fixed_source = dedent(
            """\
            ---
            key: value value value
              value value value value
              value value value value
              value value value value
              value value value value
              value value value value
              value value value value
              value value value value
              value value value value
              value
            """
        )
        config = YamlfixConfig()
        config.line_length = 20

        result = fix_code(source, config)

        assert result == fixed_source

    @pytest.mark.parametrize("none_representation", none_representations)
    def test_none_representation_config(self, none_representation: str) -> None:
        """Make `none` value representations configurable."""
        source = dedent(
            """\
            none1:
            none2: null
            none3: Null
            none4: NULL
            none5: ~
            """
        )
        fixed_none_representation = f" {none_representation}"
        if none_representation == "":
            fixed_none_representation = ""
        fixed_source = dedent(
            f"""\
            ---
            none1:{fixed_none_representation}
            none2:{fixed_none_representation}
            none3:{fixed_none_representation}
            none4:{fixed_none_representation}
            none5:{fixed_none_representation}
            """
        )
        config = YamlfixConfig()
        config.none_representation = none_representation

        result = fix_code(source, config)

        assert result == fixed_source

    def test_preserve_quotes_config(self) -> None:
        """Make it configurable. That quotes are preserved"""
        source = dedent(
            """\
            ---
            str_key1: "value"
            str_key2: 'value'
            str_key3: value
            """
        )
        config = YamlfixConfig()
        config.preserve_quotes = True

        result = fix_code(source, config)

        assert result == source

    @pytest.mark.parametrize("quote_representation", quote_representations)
    def test_quote_all_keys_and_values_config(self, quote_representation: str) -> None:
        """Quote all keys and values with configurable quote representation."""
        source = dedent(
            """\
            none_key: null
            bool_key: true
            int_key: 1
            str_key1: "value"
            str_key2: 'value'
            str_key3: value
            str_multiline: |
              value
              value
            complex_key:
              complex_key2: value
              list:
                - item1
                - item2
              complex_list:
                - item1
                - complex_item:
                    key: value
            """
        )
        quote = quote_representation
        fixed_source = dedent(
            f"""\
            ---
            {quote}none_key{quote}:
            {quote}bool_key{quote}: true
            {quote}int_key{quote}: 1
            {quote}str_key1{quote}: {quote}value{quote}
            {quote}str_key2{quote}: {quote}value{quote}
            {quote}str_key3{quote}: {quote}value{quote}
            {quote}str_multiline{quote}: |
              value
              value
            {quote}complex_key{quote}:
              {quote}complex_key2{quote}: {quote}value{quote}
              {quote}list{quote}: [{quote}item1{quote}, {quote}item2{quote}]
              {quote}complex_list{quote}:
                - {quote}item1{quote}
                - {quote}complex_item{quote}:
                    {quote}key{quote}: {quote}value{quote}
            """
        )
        config = YamlfixConfig()
        config.quote_representation = quote_representation
        config.quote_keys_and_basic_values = True

        result = fix_code(source, config)

        assert result == fixed_source

    @pytest.mark.parametrize("quote_representation", quote_representations)
    def test_quote_values_config(self, quote_representation: str) -> None:
        """Quote only scalar values with configurable quote representation."""
        source = dedent(
            """\
            none_key: null
            bool_key: true
            int_key: 1
            str_key1: "value"
            str_key2: 'value'
            str_key3: value
            str_multiline: |
              value
              value
            complex_key:
              complex_key2: value
              list:
                - item1
                - item2
              complex_list:
                - item1
                - complex_item:
                    key: 'value?'
            """
        )
        quote = quote_representation
        fixed_source = dedent(
            f"""\
            ---
            none_key:
            bool_key: true
            int_key: 1
            str_key1: {quote}value{quote}
            str_key2: {quote}value{quote}
            str_key3: {quote}value{quote}
            str_multiline: |
              value
              value
            complex_key:
              complex_key2: {quote}value{quote}
              list: [{quote}item1{quote}, {quote}item2{quote}]
              complex_list:
                - item1
                - complex_item:
                    key: {quote}value?{quote}
            """
        )
        config = YamlfixConfig()
        config.quote_representation = quote_representation
        config.quote_basic_values = True

        result = fix_code(source, config)

        assert result == fixed_source

    @pytest.mark.parametrize("quote_representation", quote_representations)
    def test_quote_all_keys_and_values_config_and_preserve_quotes(
        self, quote_representation: str
    ) -> None:
        """Quote all keys and values with configurable quote representation. \
           `quote_keys_and_basic_values` in combination with `preserve_quotes`"""
        source = dedent(
            """\
            none_key: null
            bool_key: true
            int_key: 1
            str_key1: "value"
            str_key2: 'value'
            str_key3: value
            str_multiline: |
              value
              value
            complex_key:
              complex_key2: value
              list:
                - item1
                - item2
              complex_list:
                - item1
                - complex_item:
                    key: value
            """
        )
        quote = quote_representation
        fixed_source = dedent(
            f"""\
            ---
            {quote}none_key{quote}:
            {quote}bool_key{quote}: true
            {quote}int_key{quote}: 1
            {quote}str_key1{quote}: "value"
            {quote}str_key2{quote}: 'value'
            {quote}str_key3{quote}: {quote}value{quote}
            {quote}str_multiline{quote}: |
              value
              value
            {quote}complex_key{quote}:
              {quote}complex_key2{quote}: {quote}value{quote}
              {quote}list{quote}: [{quote}item1{quote}, {quote}item2{quote}]
              {quote}complex_list{quote}:
                - {quote}item1{quote}
                - {quote}complex_item{quote}:
                    {quote}key{quote}: {quote}value{quote}
            """
        )
        config = YamlfixConfig()
        config.quote_representation = quote_representation
        config.quote_keys_and_basic_values = True
        config.preserve_quotes = True

        result = fix_code(source, config)

        assert result == fixed_source

    @pytest.mark.parametrize("quote_representation", quote_representations)
    def test_quote_values_config_and_preserve_quotes(
        self, quote_representation: str
    ) -> None:
        """Quote only scalar values with configurable quote representation. \
           `quote_basic_values` in combination with `preserve_quotes`"""
        source = dedent(
            """\
            none_key: null
            bool_key: true
            int_key: 1
            str_key1: "value"
            str_key2: 'value'
            str_key3: value
            str_multiline: |
              value
              value
            complex_key:
              complex_key2: value
              list:
                - item1
                - item2
              complex_list:
                - item1
                - complex_item:
                    key: 'value?'
            """
        )
        quote = quote_representation
        fixed_source = dedent(
            f"""\
            ---
            none_key:
            bool_key: true
            int_key: 1
            str_key1: "value"
            str_key2: 'value'
            str_key3: {quote}value{quote}
            str_multiline: |
              value
              value
            complex_key:
              complex_key2: {quote}value{quote}
              list: [{quote}item1{quote}, {quote}item2{quote}]
              complex_list:
                - item1
                - complex_item:
                    key: 'value?'
            """
        )
        config = YamlfixConfig()
        config.quote_representation = quote_representation
        config.quote_basic_values = True
        config.preserve_quotes = True

        result = fix_code(source, config)

        assert result == fixed_source

    def test_sequence_flow_style_config(self) -> None:
        """Make inline list style 'flow-style' configurable."""
        source = dedent(
            """\
            list:
              - item
              - item
            list2: [item, item]
            """
        )
        fixed_source = dedent(
            """\
            ---
            list: [item, item]
            list2: [item, item]
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.FLOW_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_sequence_block_style_config(self) -> None:
        """Make multi-line list style 'block-style' configurable."""
        source = dedent(
            """\
            list:
              - item
              - item
            list2: [item, item]
            """
        )
        fixed_source = dedent(
            """\
            ---
            list:
              - item
              - item
            list2:
              - item
              - item
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.BLOCK_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_sequence_keep_style_config(self) -> None:
        """Make it configurable, that the list style is not changed and keeps\
            the original flow- or block-style for sequences."""
        source = dedent(
            """\
            ---
            list:
              - item
              - item
            list2: [item, item]
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.KEEP_STYLE

        result = fix_code(source, config)

        assert result == source

    def test_sequence_block_style_enforcement_for_lists_with_comments(self) -> None:
        """Fall back to multi-line list style 'block-style' if list contains comments,\
            even if flow-style is selected."""
        source = dedent(
            """\
            list: # List comment
              # Comment 1
              - item
              # Comment 2
              - item
            list2: # List 2 Comment
              - item
              - item
            list3: # List 3 Comment
              - item with long description
              - item with long description
              - item with long description
              - item with long description
            """
        )
        fixed_source = dedent(
            """\
            ---
            list:  # List comment
              # Comment 1
              - item
              # Comment 2
              - item
            list2: [item, item]  # List 2 Comment
            list3:  # List 3 Comment
              - item with long description
              - item with long description
              - item with long description
              - item with long description
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.FLOW_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_sequence_block_style_enforcement_for_lists_with_non_scalar_values(
        self,
    ) -> None:
        """Fall back to multi-line list style 'block-style' if list contains non-scalar\
            values, like other lists or dicts, even if flow-style is selected."""
        source = dedent(
            """\
            list:
              - nested_list:
                  - item
              - item
            list2:
              - item
              - nested_dict:
                  key: value
            list3:
              - item
              - item
            """
        )
        fixed_source = dedent(
            """\
            ---
            list:
              - nested_list: [item]
              - item
            list2:
              - item
              - nested_dict:
                  key: value
            list3: [item, item]
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.FLOW_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_sequence_block_style_enforcement_for_lists_longer_than_line_length(
        self,
    ) -> None:
        """Fall back to multi-line list style 'block-style' if list would be longer than\
            line_length, even if flow-style is selected."""
        source = dedent(
            """\
            looooooooooooooooooooooooooooooooooooongKey:
              - item
            list:
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
            list2:
              - item
              - item
              - item
            """
        )
        fixed_source = dedent(
            """\
            ---
            looooooooooooooooooooooooooooooooooooongKey:
              - item
            list:
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
              - loooooooooooooooooooongItem
            list2: [item, item, item]
            """
        )
        config = YamlfixConfig()
        config.line_length = 40
        config.sequence_style = YamlNodeStyle.FLOW_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_sequence_flow_style_with_trailing_newlines(self) -> None:
        """Correct flow-style lists that have trailing newlines.

        Without this fix the following block-style list:
        ```
        list:
          - item
          - item



        key: value
        ```

        is converted to this weird-looking flow-style list:
        ```
        list: [item, item



        ]
        key: value
        ```

        instead of
        ```
        list: [item, item]



        key: value
        ```
        """
        source = dedent(
            """\
            list:
              - item
              - item



            key: value
            """
        )
        fixed_source = dedent(
            """\
            ---
            list: [item, item]
            key: value
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.FLOW_STYLE

        result = fix_code(source, config)

        assert result == fixed_source

    def test_empty_list_inline_comment_indentation(self) -> None:
        """Check if inline comment is preserved for empty lists with comments."""
        source = dedent(
            """\
            ---
            indented:
              key: value
              list1: [value]  # comment with value
              list2: []
              list3: []  # comment on the same line as empty list
              map:
                anotherKey: anotherValue
            """
        )
        config = YamlfixConfig()
        config.sequence_style = YamlNodeStyle.FLOW_STYLE

        result = fix_code(source, config)

        assert result == source

    def test_section_whitelines(self) -> None:
        """Check if section whitelines are preserved."""
        source = dedent(
            # pylint: disable=C0303
            """\
            ---

            begin_section:
              key: value
            key1: value

            key2: value

            happy_path_section:
              key1: value

              key2: value
              nested_dict:
                nested_key: value

            # Comment 1
            # Comment 2
            comment_section:
                key: value
            key3: value
            key4: value

            key5: value
            close_section:
                key: value



            """  # noqa: W291
        )
        fixed_source = dedent(
            """\
            ---
            begin_section:
              key: value

            key1: value
            key2: value

            happy_path_section:
              key1: value
              key2: value
              nested_dict:
                nested_key: value


            # Comment 1
            # Comment 2
            comment_section:
              key: value

            key3: value
            key4: value
            key5: value

            close_section:
              key: value
            """
        )
        config = YamlfixConfig()
        config.section_whitelines = 1
        config.comments_whitelines = 2

        result = fix_code(source, config)

        assert result == fixed_source

    def test_section_whitelines_begin_no_explicit_start(self) -> None:
        """Check that no whitelines are added at start of file when explicit start \
            is not applied."""
        source = dedent(
            # pylint: disable=C0303
            """\
            begin_section:
              key: value
            """  # noqa: W291
        )
        fixed_source = dedent(
            """\
            begin_section:
              key: value
            """
        )
        config = YamlfixConfig()
        config.section_whitelines = 1
        config.comments_whitelines = 2
        config.explicit_start = False

        result = fix_code(source, config)

        assert result == fixed_source

    def test_whitelines_collapsed(self) -> None:
        """Checks that whitelines are collapsed by default."""
        source = dedent(
            """\
            key: value

            dict:
              key: value
              nested_dict:
                - key: value
                  key2: value2

                - key: value
            """
        )
        fixed_source = dedent(
            """\
            ---
            key: value
            dict:
              key: value
              nested_dict:
                - key: value
                  key2: value2
                - key: value
            """
        )
        config = YamlfixConfig()

        result = fix_code(source, config)

        assert result == fixed_source

    def test_whitelines_adjusted_to_value(self) -> None:
        """Checks that amount of whitelines are in line with the config value."""
        source = dedent(
            """\
            key: value

            dict:
              key: value


              nested_list:
                - key: value
                  key2: value2

                - key: value
            """
        )
        fixed_source = dedent(
            """\
            ---
            key: value

            dict:
              key: value

              nested_list:
                - key: value
                  key2: value2

                - key: value
            """
        )
        config = YamlfixConfig()
        config.whitelines = 1

        result = fix_code(source, config)

        assert result == fixed_source

    def test_whitelines_higher_than_secion_whitelines(self) -> None:
        """Checks that amount of whitelines are in line with the config values."""
        source = dedent(
            # pylint: disable=C0303
            """\
            ---
            begin_section:
              key: value
            key1: value

            key2: value
            happy_path_section:
              key1: value

              key2: value
              nested_dict:
                nested_key: value

            # Comment 1
            # Comment 2
            comment_section:
                key: value
            key3: value
            key4: value

            key5: value
            close_section:
                key: value



            """  # noqa: W291
        )
        fixed_source = dedent(
            """\
            ---
            begin_section:
              key: value


            key1: value

            key2: value


            happy_path_section:
              key1: value

              key2: value
              nested_dict:
                nested_key: value

            # Comment 1
            # Comment 2
            comment_section:
              key: value


            key3: value
            key4: value

            key5: value


            close_section:
              key: value
            """
        )
        config = YamlfixConfig()
        config.whitelines = 1
        config.section_whitelines = 2

        result = fix_code(source, config)

        assert result == fixed_source

    def test_enforcing_flow_style_together_with_adjustable_newlines(self) -> None:
        """Checks that transforming block style sequences to flow style together with
        newlines adjusting produces correct result.
        """
        source = dedent(
            """\
            ---
            dict:
              nested_dict:
                key: value
                key2:
                  - list_item


              nested_dict2:
                key: value
            """
        )
        fixed_source = dedent(
            """\
            ---
            dict:
              nested_dict:
                key: value
                key2: [list_item]

              nested_dict2:
                key: value
            """
        )
        config = YamlfixConfig()
        config.whitelines = 1
        config.sequence_style = YamlNodeStyle.FLOW_STYLE

        result = fix_code(source, config)

        assert result == fixed_source
